require 'elasticord/client/array_with_meta_data'

module Elasticord
  module Client
    class SearchBuilder
      extend Forwardable

      PAGE_DEFAULT = 1
      PER_PAGE_DEFAULT = 10

      def initialize(elastic_search_client, entity, index, type)
        @elastic_search_client, @entity, @index, @type =
          elastic_search_client, entity, index, type
      end

      attr_reader :index, :type

      def_delegators :results, :current_page, :per_page, :size, :from

      def page(new_page)
        results.current_page = new_page

        body_params[:from] = results.from

        self
      end

      def per(new_per_page)
        results.per_page = new_per_page

        body_params[:size] = results.size
        body_params[:from] = results.from

        self
      end

      def load(body = {})
        begin
          elastic_results = execute_search(body)

          results.total_results = elastic_results['total']

          elastic_results['hits'].map do |result|
            results.push entity.new(result['_source'])
          end
        rescue Elasticsearch::Transport::Transport::Errors::NotFound
          nil
        end

        results
      end

      # def ft(full_text_term)
      #   return self unless full_text_term
      #
      #   params = (body_params[:query][:bool][:must] ||= [])
      #
      #   # params[:must] << { match: { title: full_text_term } }
      # end

      def ransack(ransack_filters = {})
        return self unless ransack_filters

        params = body_params[:query][:bool]

        ransack_filters.each do |key, value|
          if key['_eq_any']
            add_should_clause(params, key, value)
          elsif key['_eq']
            add_must_clause(params, key, value)
          end
        end

        self
      end

      def body_params
        @body_params ||= { size: size, from: from, query: { bool: {} } }
      end

      protected

      attr_reader :elastic_search_client, :entity

      def results
        @results ||= ArrayWithMetaData.new \
          PAGE_DEFAULT, PER_PAGE_DEFAULT
      end

      def execute_search(body)
        elastic_results = elastic_search_client.search \
          index: index, type: type, body: body_params.merge(body)

        hits = elastic_results['hits']

        (hits && hits['total'].to_i > 0) ? hits : { 'hits' => [], 'total' => 0 }
      end

      def add_must_clause(params, key, value)
        params[:must] ||= []

        key = key.to_s.split('_eq').first.to_sym

        add_term(params[:must], key, value)
      end

      def add_should_clause(params, key, values)
        params[:boost] = 1.0
        params[:should] ||= []
        params[:minimum_should_match] = 1

        key = key.to_s.split('_eq_any').first.to_sym

        [*values].each { |value| add_term(params[:should], key, value) }
      end

      private

      def add_term(params, key, value)
        value = value.downcase if value.is_a? String

        params.push({ term: { key => value } })
      end
    end
  end
end
